import pandas
import h5py

from sparsedat import Sparse_Data_Table


def write_pandas_csv(data_frame, file_path):
    pandas.DataFrame(data_frame)\
        .to_csv(file_path, sep=',', encoding='utf-8', chunksize=1000)


def read_pandas_csv(file_path):
    return pandas.read_csv(file_path, sep=",", header=0, index_col=0)


def convert_h5_to_sdt(
        h5_file_path,
        SDT_file_path,
        cells_as_rows=True
):
    h5_file = h5py.File(h5_file_path)

    cellranger_version = 2

    if "matrix" in h5_file:
        cellranger_version = 3

    matrix_name = None

    if cellranger_version == 2:
        for key, value in h5_file.items():
            matrix_name = key
            break
    else:
        matrix_name = "matrix"

    sdt = Sparse_Data_Table()

    data = h5_file[matrix_name]["data"][()]
    indices = h5_file[matrix_name]["indices"][()]
    indptr = h5_file[matrix_name]["indptr"][()]

    if indices[0] > indices[1]:
        for column_index in range(len(indptr) - 1):
            data[indptr[column_index]:indptr[column_index + 1]] = \
                data[indptr[column_index]:indptr[column_index + 1]][::-1]
            indices[indptr[column_index]:indptr[column_index + 1]] = \
                indices[indptr[column_index]:indptr[column_index + 1]][::-1]
    else:
        for column_index in range(len(indptr) - 1):
            data[indptr[column_index]:indptr[column_index + 1]] = \
                data[indptr[column_index]:indptr[column_index + 1]]
            indices[indptr[column_index]:indptr[column_index + 1]] = \
                indices[indptr[column_index]:indptr[column_index + 1]]

    sdt.from_sparse_column_entries(
        (
            data,
            indices,
            indptr
        ),
        h5_file[matrix_name]["shape"][0],
        h5_file[matrix_name]["shape"][1]
    )

    if cellranger_version == 2:
        gene_names = [x.decode("UTF-8")
                      for x in list(h5_file[matrix_name]["gene_names"])]
        gene_ids = [x.decode("UTF-8")
                    for x in list(h5_file[matrix_name]["gene_ids"])]
    else:
        gene_names = [x.decode("UTF-8")
                      for x in list(h5_file[matrix_name]["features"]["name"])]
        gene_ids = [x.decode("UTF-8")
                    for x in list(h5_file[matrix_name]["features"]["id"])]

    gene_name_indices = {}

    disambiguated_gene_names = []

    for gene_index, gene in enumerate(gene_names):
        if gene not in gene_name_indices:
            gene_name_indices[gene] = [gene_ids[gene_index]]
        else:
            gene_name_indices[gene].append(gene_ids[gene_index])

    for gene_index, gene in enumerate(gene_names):

        if len(gene_name_indices[gene]) > 1:
            # Figure out which gene this is, as sorted by id
            this_gene_id = gene_ids[gene_index]
            duplicate_gene_index = sorted(gene_name_indices[gene]).index(
                this_gene_id)
            disambiguated_gene_name = "%s_%i" % (gene, duplicate_gene_index + 1)
            disambiguated_gene_names.append(disambiguated_gene_name)
        else:
            disambiguated_gene_names.append(gene)

    sdt.column_names = [x.decode("utf-8") for x in
                        h5_file[matrix_name]["barcodes"][()]]
    sdt.row_names = disambiguated_gene_names

    # Transpose so that rows are cells
    if cells_as_rows:
        sdt.transpose()

    sdt.save(SDT_file_path)



import os

from pepars.fileio import fileio
from pepars.utils import DNA as DNA_utils
from pepars.utils import Sequence_Trie


def get_chromium_barcodes(chromium_version=3):

    URL_prefix = "https://caltech.box.com/shared/static/"

    if chromium_version == 2:
        file_name = "cell-barcodes.10x.737K-august-2016.txt"
        remote_file_URL = URL_prefix + "du9v3pfdammy3u1iwgal8fin495u6kw9.txt"
    elif chromium_version == 3:
        file_name = "3M-february-2018.txt"
        remote_file_URL = URL_prefix + "evjhzh50rk6zk2jtjpfg3myddxxmnqn5.txt"
    else:
        raise NotImplementedError("Only support versions 2 and 3 of Chromium")

    valid_barcodes_file_path = os.path.join("data", file_name)

    fileio.download_remote_file(
        file_URL=remote_file_URL,
        file_path=valid_barcodes_file_path
    )

    valid_barcode_trie = Sequence_Trie(
        DNA_utils.NUCLEOTIDE_INDEX_MAP,
        allow_invalid=True)

    with open(valid_barcodes_file_path) as valid_barcodes_file:
        for line in valid_barcodes_file.readlines():
            valid_barcode_trie.add(line.strip())

    return valid_barcode_trie



import pandas

from pepars.utils import FASTQ as FASTQ_utils
from pepars.utils import DNA as DNA_utils
from pepars.utils import Sequence_Trie
from pepars.analysis import DNA as DNA_analysis


def get_cell_tag_counts(
        FASTQ_file_set,
        valid_cell_barcode_trie=None,
        valid_tags=None,
        index_sequence=None,
        maximum_tag_distance=0,
        maximum_index_distance=0,
        maximum_cell_barcode_distance=0,
        quality_threshold=None,
        collapse_UMIs=True,
        error_correct=False,
        cell_barcode_range=(0, 16),
        cell_barcode_read_number=1,
        tag_read_number=2,
        tag_range=(0, 8),
        UMI_range=(16, 28),
        UMI_read_number=1,
        output_frequency=1e5
):

    tag_start_index = tag_range[0]
    tag_end_index = tag_range[1]
    tag_length = tag_end_index - tag_start_index

    UMI_start_index = UMI_range[0]
    UMI_end_index = UMI_range[1]
    UMI_length = UMI_end_index - UMI_start_index

    cell_barcode_start_index = cell_barcode_range[0]
    cell_barcode_end_index = cell_barcode_range[1]
    cell_barcode_length = cell_barcode_end_index - cell_barcode_start_index

    if valid_tags:
        valid_tag_trie = Sequence_Trie(DNA_utils.NUCLEOTIDE_INDEX_MAP,
                                       allow_invalid=True)

        for tag in valid_tags:
            valid_tag_trie.add(tag)
    else:
        valid_tag_trie = None

    cell_UMI_tag_counts = {}
    num_reads_passing_quality_threshold = 0
    num_reads_matching_index = 0
    num_reads_matching_tag = 0
    num_tag_reads_matching_cell = 0
    num_reads = 0

    if quality_threshold:
        FASTQ_file_set_iterator = FASTQ_file_set.get_sequence_quality_iterator()
    else:
        FASTQ_file_set_iterator = FASTQ_file_set.get_sequence_iterator()

    for read in FASTQ_file_set_iterator:

        num_reads += 1

        if output_frequency and num_reads % output_frequency == 0:
            print("Read %i reads" % num_reads)

        if quality_threshold:

            passes_quality_score = True

            index_quality_sequence = read[1][0]

            index_quality_scores = FASTQ_utils.convert_quality_string_to_quality_score(
                index_quality_sequence)

            for quality_score in index_quality_scores:
                if quality_score < quality_threshold:
                    passes_quality_score = False
                    break

            if not passes_quality_score:
                continue

            cell_barcode_quality_sequence = read[1][cell_barcode_read_number][
                                            cell_barcode_start_index:cell_barcode_end_index]

            cell_barcode_quality_scores = FASTQ_utils.convert_quality_string_to_quality_score(
                cell_barcode_quality_sequence)

            for quality_score in cell_barcode_quality_scores:
                if quality_score < quality_threshold:
                    passes_quality_score = False
                    break

            if not passes_quality_score:
                continue

            UMI_quality_sequence = read[1][UMI_read_number][
                                   UMI_start_index:UMI_end_index]

            UMI_quality_scores = FASTQ_utils.convert_quality_string_to_quality_score(
                UMI_quality_sequence)

            for quality_score in UMI_quality_scores:
                if quality_score < quality_threshold:
                    passes_quality_score = False
                    break

            if not passes_quality_score:
                continue

            tag_quality_sequence = read[1][tag_read_number][
                                   tag_start_index:tag_end_index]

            tag_quality_scores = FASTQ_utils.convert_quality_string_to_quality_score(
                tag_quality_sequence)

            for quality_score in tag_quality_scores:
                if quality_score < quality_threshold:
                    passes_quality_score = False
                    break

            if not passes_quality_score:
                continue

            sequences = read[0]
        else:
            sequences = read

        num_reads_passing_quality_threshold += 1

        if index_sequence:
            if not DNA_utils.is_template_match(
                    index_sequence,
                    sequences[0],
                    allowable_distance=maximum_index_distance):
                continue

        num_reads_matching_index += 1

        tag_sequence = sequences[tag_read_number][tag_start_index:tag_end_index]

        if error_correct:
            cell_barcode = sequences[cell_barcode_read_number][
                           cell_barcode_start_index:cell_barcode_end_index]
            UMI = sequences[UMI_read_number][UMI_start_index:UMI_end_index]
        else:
            if valid_tag_trie:
                if not valid_tag_trie.find(tag_sequence):
                    if maximum_tag_distance == 0:
                        continue
                    else:
                        possible_tags = DNA_analysis.find_all_sequences_of_distance_n(
                            tag_sequence,
                            valid_tag_trie,
                            n=maximum_tag_distance)

                        if len(possible_tags) != 1:
                            continue

                        tag_sequence = possible_tags[0]

            num_reads_matching_tag += 1

            cell_barcode = sequences[cell_barcode_read_number][
                           cell_barcode_start_index:cell_barcode_end_index]

            if valid_cell_barcode_trie:
                if not valid_cell_barcode_trie.find(cell_barcode):
                    if maximum_cell_barcode_distance == 0:
                        continue
                    else:
                        possible_cell_barcodes = DNA_analysis.find_all_sequences_of_distance_n(
                            cell_barcode,
                            valid_cell_barcode_trie,
                            n=maximum_cell_barcode_distance)

                        if len(possible_cell_barcodes) != 1:
                            continue

                        cell_barcode = possible_cell_barcodes[0]

            num_tag_reads_matching_cell += 1

            UMI = sequences[UMI_read_number][UMI_start_index:UMI_end_index]

        cell_UMI_tag = cell_barcode + UMI + tag_sequence

        if cell_UMI_tag not in cell_UMI_tag_counts:
            cell_UMI_tag_counts[cell_UMI_tag] = 1
        else:
            cell_UMI_tag_counts[cell_UMI_tag] += 1

    cell_UMI_tag_counts_list = [(cell_UMI_tag, count) for cell_UMI_tag, count in
                                cell_UMI_tag_counts.items()]

    if error_correct:

        cell_UMI_tag_counts_list = DNA_analysis.collapse_similar_sequences(
            cell_UMI_tag_counts_list)

        filtered_cell_UMI_tag_counts = {}

        for cell_UMI_tag, count in cell_UMI_tag_counts_list:

            tag_sequence = cell_UMI_tag[cell_barcode_length + UMI_length:]

            if valid_tag_trie:
                if not valid_tag_trie.find(tag_sequence):
                    if maximum_tag_distance == 0:
                        continue
                    else:
                        possible_tags = DNA_analysis.find_all_sequences_of_distance_n(
                            tag_sequence,
                            valid_tag_trie,
                            n=maximum_tag_distance)

                        if len(possible_tags) != 1:
                            continue

                        tag_sequence = possible_tags[0]

            num_reads_matching_tag += count

            cell_barcode = cell_UMI_tag[0:cell_barcode_length]

            if valid_cell_barcode_trie:
                if not valid_cell_barcode_trie.find(cell_barcode):
                    if maximum_cell_barcode_distance == 0:
                        continue
                    else:
                        possible_cell_barcodes = DNA_analysis.find_all_sequences_of_distance_n(
                            cell_barcode,
                            valid_cell_barcode_trie,
                            n=maximum_cell_barcode_distance)

                        if len(possible_cell_barcodes) != 1:
                            continue

                        cell_barcode = possible_cell_barcodes[0]

            num_tag_reads_matching_cell += count

            UMI = cell_UMI_tag[
                  cell_barcode_length:cell_barcode_length + UMI_length]

            cell_UMI_tag = cell_barcode + UMI + tag_sequence

            if cell_UMI_tag not in filtered_cell_UMI_tag_counts:
                filtered_cell_UMI_tag_counts[cell_UMI_tag] = 1
            else:
                filtered_cell_UMI_tag_counts[cell_UMI_tag] += 1

        cell_UMI_tag_counts_list = [(cell_UMI_tag, count) for
                                    cell_UMI_tag, count in
                                    filtered_cell_UMI_tag_counts.items()]

    # Establish our tag list, either from the given list, or as deduced from the data
    if not valid_tags:
        valid_tags = set({})

        for cell_UMI_tag in cell_UMI_tag_counts_list:
            tag = cell_UMI_tag[0][cell_barcode_length + UMI_length:]
            valid_tags.add(tag)

    cell_tag_counts = {}

    tag_indices = {}

    tag_index = 0
    sorted_tag_list = []

    for tag in sorted(valid_tags):
        tag_indices[tag] = tag_index
        tag_index += 1
        sorted_tag_list.append(tag)

    for sequence, count in cell_UMI_tag_counts_list:

        cell_barcode = sequence[0:cell_barcode_length]
        tag = sequence[cell_barcode_length + UMI_length:]

        try:
            tag_index = tag_indices[tag]

            if cell_barcode not in cell_tag_counts:
                cell_tag_counts[cell_barcode] = [0 for tag in valid_tags]

            if collapse_UMIs:
                cell_tag_counts[cell_barcode][tag_index] += 1
            else:
                cell_tag_counts[cell_barcode][tag_index] += count
        except ValueError:
            pass

    cell_tag_counts_df = pandas.DataFrame.from_dict(cell_tag_counts,
                                                    orient="index")
    cell_tag_counts_df.columns = sorted_tag_list

    print("Num reads: %i" % num_reads)
    print("Reads passing quality threshold: %i/%i (%.2f%%)" % (
        num_reads_passing_quality_threshold, num_reads,
        num_reads_passing_quality_threshold / num_reads * 100))
    print("Reads matching index: %i/%i (%.2f%%)" % (
        num_reads_matching_index, num_reads_passing_quality_threshold,
        num_reads_matching_index / num_reads_passing_quality_threshold * 100))
    print("Reads matching tag: %i/%i (%.2f%%)" % (
        num_reads_matching_tag, num_reads_matching_index,
        num_reads_matching_tag / num_reads_matching_index * 100))
    print("Tag reads matching cell: %i/%i (%.2f%%)" % (
        num_tag_reads_matching_cell, num_reads_matching_tag,
        num_tag_reads_matching_cell / num_reads_matching_tag * 100))

    return cell_tag_counts_df



from setuptools import setup

setup(
    name="scrapi",
    version="0.3.1",
    packages=[
        "scrap",
        "scrap.tags",
        "scrap.chromium",
        "scrap.plotting",
        "scrap.dataset",
        "scrap.utils"
    ],
    entry_points={
        "console_scripts": [
            "scrap-preprocess=scrap.dataset.preprocess:preprocess",
            "scrap-init-ws=scrap.dataset.initialize:initialize"
        ]
    },
    install_requires=[
        "pandas",
        "pepars",
        "sparsedat",
        "numpy",
        "plotly",
        "h5py",
        "sklearn",
        "scipy",
        "statsmodels"
    ]
)



import os
import argparse

from sparsedat import wrappers as sparsedat_wrappers

from .. import utils as scrap_utils
from .Gene_Expression_Dataset import Gene_Expression_Dataset


def get_arguments():

    argparser = argparse.ArgumentParser(
        description="Initialize a SCRAP workspace")
    argparser.add_argument("--workspace_path", "-w",
                           help="Path to the workspace", default=None)
    argparser.add_argument("--h5_file_path", "-f",
                           help="Path to the H5 file", default=None)
    argparser.add_argument("--csv_file_path", "-c",
                           help="Path to a CSV file", default=None)
    argparser.add_argument("--mtx_directory_path", "-m",
                           help="Path to an mtx directory", default=None)
    argparser.add_argument("--sdt_file_path", "-s",
                           help="Path to an SDT file", default=None)

    args = argparser.parse_args()

    return args


def initialize():

    args = get_arguments()

    if args.workspace_path is None:
        workspace_path = os.path.join(os.getcwd(), "workspace")
    else:
        workspace_path = args.workspace_path

    if os.path.isfile(workspace_path):
        raise ValueError("Cannot use file as a workspace")

    if not os.path.isdir(workspace_path):
        os.makedirs(workspace_path, exist_ok=True)

    if args.h5_file_path is None and \
            args.csv_file_path is None and \
            args.mtx_directory_path is None and \
            args.sdt_file_path is None:
        raise ValueError("Must specify path to file with -f, -c, -m, or -s")

    raw_transcript_counts_file_path = os.path.join(
        workspace_path, "source_barcode_transcript_counts.sdt")

    if args.sdt_file_path is not None:
        raw_transcript_counts_file_path = args.sdt_file_path
    elif args.h5_file_path is not None:
        scrap_utils.convert_h5_to_sdt(
            args.h5_file_path,
            raw_transcript_counts_file_path)
    elif args.csv_file_path is not None:
        raise NotImplementedError("Haven't implemented csv to sdt")
    else:
        sdt = sparsedat_wrappers.load_mtx(
            os.path.join(args.mtx_directory_path, "features.tsv"),
            os.path.join(args.mtx_directory_path, "barcodes.tsv"),
            os.path.join(args.mtx_directory_path, "matrix.mtx")
        )
        sdt.save(raw_transcript_counts_file_path)

    seed_matrices_file_paths = [raw_transcript_counts_file_path]

    Gene_Expression_Dataset.initialize_dataset(
        workspace_path,
        seed_matrices_file_path=seed_matrices_file_paths
    )
